﻿using Connected.Common.Types.TaxRates;
using Connected.Entities;
using Connected.Entities.Storage;
using Connected.Notifications.Events;
using Connected.ServiceModel;
using Connected.ServiceModel.Transactions;

using Moq;
using static Connected.Common.Types.TaxRates.TaxRateOps;

namespace Common.Types.TaxRates.Ops;

public class DeleteTests
{
	[TestClass]
	public class OnInvoke
	{
		[TestMethod("Delete invoke calls update with state as deleted")]
		public async Task Delete_Invoke_CallsUpdateWithStateDeleted()
		{
			/*
			 * Setup
			 */
			var expectedValue = State.Deleted;
			var actualValue = (State)0;
			/* 
			 * Required services
			 */
			var instanceFaker = new InstanceFaker<Delete>();

			var databaseContextProviderFake = instanceFaker.GetMock<IStorageProvider>()!;

			_ = databaseContextProviderFake
				.Setup(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()))
				.Callback<TaxRate>(e => actualValue = e.State)
				.ReturnsAsync(new TaxRate { Id = 1 });
			/*
			 * Arguments play no role for this test
			 */
			var args = new PrimaryKeyArgs<int>(1);
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			await serviceFunction.Invoke(args);
			/* 
			 * Assert
			 */
			databaseContextProviderFake.Verify(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()), Times.Exactly(1));
			Assert.AreEqual(expectedValue, actualValue);
		}

		[TestMethod("Delete invoke correctly maps values to entity")]
		public async Task Delete_Invoke_MapsArgumentsCorrectly()
		{
			/* 
			 * Setup
			 */
			var expectedId = 1;
			var actualValue = int.MinValue;
			/* 
			 * Required services
			 */
			var instanceFaker = new InstanceFaker<Delete>();

			var databaseContextProviderFake = instanceFaker.GetMock<IStorageProvider>()!;

			_ = databaseContextProviderFake
				.Setup(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()))
				.Callback<TaxRate>(e => actualValue = e.Id)
				.ReturnsAsync(new TaxRate { Id = expectedId });
			/*
			 * Arguments play no role for this test
			 */
			var args = new PrimaryKeyArgs<int>(expectedId);
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			await serviceFunction.Invoke(args);
			/* 
			 * Assert
			 */
			databaseContextProviderFake.Verify(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()), Times.Exactly(1));
			Assert.AreEqual(expectedId, actualValue);
		}
	}

	[TestClass]
	public class OnCommit
	{
		[TestMethod("Delete commit invokes deleted event exactly once")]
		public async Task Delete_Commit_InvokedDeleted_ExactlyOnce()
		{
			/* 
			 * Setup
			 */
			var instanceFaker = new InstanceFaker<Delete>();

			var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

			_ = taxRateCacheFake.Setup(e => e.Refresh(It.IsAny<int>()));

			var eventServiceFake = instanceFaker.GetMock<IEventService>()!;

			_ = eventServiceFake.SetupEvent<ITaxRateService, int>(ServiceEvents.Deleted);
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;

			serviceFunction.SetArguments(new PrimaryKeyArgs<int>(1));

			await ((ITransactionClient)serviceFunction).Commit();
			/* 
			 * Assert
			 */
			eventServiceFake.VerifyEvent<ITaxRateService, int>(ServiceEvents.Deleted);
		}

		[TestMethod("Delete commit removes cache exactly once")]
		public async Task Delete_Commit_RemovesCache_ExactlyOnce()
		{
			/* 
			 * Setup
			 */
			var instanceFaker = new InstanceFaker<Delete>();

			var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

			_ = taxRateCacheFake.Setup(e => e.Remove(It.IsAny<int>()));

			var eventServiceFake = instanceFaker.GetMock<IEventService>()!;

			_ = eventServiceFake.SetupEvent<ITaxRateService, int>(ServiceEvents.Deleted);
			/*
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;

			serviceFunction.SetArguments(new PrimaryKeyArgs<int>(1));

			await ((ITransactionClient)serviceFunction).Commit();
			/* 
			 * Assert
			 */
			taxRateCacheFake.Verify(e => e.Remove(It.IsAny<int>()), Times.Exactly(1));
		}

		[TestMethod("Delete commit fires deleted event after notifying cache")]
		public async Task Delete_Commit_InvokedDeletedEvent_AfterNotifyingCache()
		{
			/* 
			 * Setup
			 * 
			 * If you receive an error about function calls not properly set up, and function 
			 * calls look ok, it's probably the sequence throwing the exception. This is a 
			 * failure in user code, not the test.
			 */
			var sequence = new MockSequence();

			var instanceFaker = new InstanceFaker<Delete>();

			var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

			_ = taxRateCacheFake
				.InSequence(sequence)
				.Setup(e => e.Remove(It.IsAny<int>()));

			var eventServiceFake = instanceFaker.GetMock<IEventService>()!;

			_ = eventServiceFake
				.InSequence(sequence)
				.SetupEvent<ITaxRateService, int>(ServiceEvents.Deleted);
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;

			serviceFunction.SetArguments(new PrimaryKeyArgs<int>(1));

			await ((ITransactionClient)serviceFunction).Commit();
			/* 
			 * Assert
			 */
			taxRateCacheFake.Verify(e => e.Remove(It.IsAny<int>()), Times.Exactly(1));
			eventServiceFake.VerifyEvent<ITaxRateService, int>(ServiceEvents.Deleted);
		}
	}
}