﻿using System.Reflection;
using Connected.Common.Types.TaxRates;
using Connected.Data;
using Connected.Entities;
using Connected.Entities.Annotations;
using Connected.Entities.Storage;
using Connected.Notifications.Events;
using Connected.ServiceModel;
using Connected.ServiceModel.Transactions;
using Moq;
using static Connected.Common.Types.TaxRates.TaxRateOps;

namespace Common.Types.TaxRates.Ops;

public class InsertTests
{
	[TestClass]
	public class OnInvoke
	{
		[TestMethod("Insert invoke returns inserted id")]
		public async Task Insert_Invoke_ReturnsInsertedId()
		{
			/* 
			 * Setup
			 */
			var expectedId = 1;
			/* 
			 * Required services
			 */
			var instanceFaker = new InstanceFaker<Insert>();

			var databaseContextProviderFake = instanceFaker.GetMock<IStorageProvider>()!;

			_ = databaseContextProviderFake
				.Setup(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()))
				.ReturnsAsync(new TaxRate { Id = expectedId });
			/*
			 * Arguments play no role for this test
			 */
			var args = new InsertTaxRateArgs { };
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			var result = await serviceFunction.Invoke(args);
			/* 
			 * Assert
			 */
			Assert.AreEqual(result, expectedId);
			databaseContextProviderFake.Verify(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()), Times.Exactly(1));
		}

		[TestMethod("Insert invoke inserts exactly one new value")]
		public async Task Insert_Invoke_InsertsNewValue_ExactlyOnce()
		{
			/* 
			 * Setup
			 */
			var expectedId = 1;
			/* 
			 * Required services
			 */
			var instanceFaker = new InstanceFaker<Insert>();

			var databaseContextProviderFake = instanceFaker.GetMock<IStorageProvider>()!;

			_ = databaseContextProviderFake
				.Setup(e => e.Open<TaxRate>().Update(It.Is<TaxRate>(e => e.State == State.New)))
				.ReturnsAsync(new TaxRate { Id = expectedId });
			/*
			 * Arguments play no role for this test
			 */
			var args = new InsertTaxRateArgs();
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			var result = await serviceFunction.Invoke(args);
			/* 
			 * Assert
			 */
			databaseContextProviderFake.Verify(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()), Times.Exactly(1));
		}

		[TestMethod("Insert invoke correctly maps values to entity")]
		public async Task Insert_Invoke_MapsArgumentsCorrectly()
		{
			/* 
			 * Setup
			 */
			var expectedEntity = new TaxRate
			{
				Name = "TaxRate",
				Rate = 1.0f,
			};
			/* 
			 * Required services
			 */
			var actualEntity = null as TaxRate;

			var instanceFaker = new InstanceFaker<Insert>();

			var databaseContextProviderFake = instanceFaker.GetMock<IStorageProvider>()!;

			_ = databaseContextProviderFake
				.Setup(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()))
				.Callback<TaxRate>(e => actualEntity = e)
				.ReturnsAsync(expectedEntity with { Id = 1 });

			var args = new InsertTaxRateArgs
			{
				Name = expectedEntity.Name,
				Rate = expectedEntity.Rate
			};
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			var result = await serviceFunction.Invoke(args);
			/* 
			 * Assert
			 */
			databaseContextProviderFake.Verify(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()), Times.Exactly(1));
			Assert.AreEqual(expectedEntity.Name, actualEntity?.Name);
			Assert.AreEqual(expectedEntity.Rate, actualEntity?.Rate);
		}

		[TestMethod("Insert invoke sets status to enabled or default")]
		public async Task Insert_Invoke_SetsStatusToEnabledOrDefault()
		{
			/*
			 * Setup
			 * Valid values are Enabled or Default (db provider sets it to enabled then)
			 */
			var expectedStatuses = new[] { Status.Enabled }.ToList();
			/*
			 * Check if the default value for the status field is enabled. If so, add 0 to the expected values list.
			 */
			var defaultValueAttributes = typeof(TaxRate).GetProperty(nameof(TaxRate.Status))!.GetCustomAttributes<DefaultAttribute>(true);

			if (defaultValueAttributes.Any())
			{
				var values = defaultValueAttributes.Select(e => (Status)e.Value!).ToList();

				if (values.Any(e => e == Status.Enabled))
					expectedStatuses.Add(0);
			}
			/*
			 * Required services
			 */
			var actualStatus = (Status)0;

			var instanceFaker = new InstanceFaker<Insert>();

			var databaseContextProviderFake = instanceFaker.GetMock<IStorageProvider>()!;

			_ = databaseContextProviderFake
				.Setup(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()))
				.Callback<TaxRate>(e => actualStatus = e.Status)
				.ReturnsAsync(new TaxRate { Id = 1 });
			/*
			 * Arguments play no role for this test
			 */
			var args = new InsertTaxRateArgs();
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			var result = await serviceFunction.Invoke(args);
			/* 
			 * Assert
			 */
			databaseContextProviderFake.Verify(e => e.Open<TaxRate>().Update(It.IsAny<TaxRate>()), Times.Exactly(1));
			Assert.IsTrue(expectedStatuses.Contains(actualStatus));
		}
	}

	[TestClass]
	public class OnCommit
	{
		[TestMethod("Insert commit invokes inserted event exactly once")]
		public async Task Insert_Commit_InvokedInserted_ExactlyOnce()
		{
			/* 
			 * Setup
			 */
			var instanceFaker = new InstanceFaker<Insert>();

			var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

			_ = taxRateCacheFake.Setup(e => e.Refresh(It.IsAny<int>()));

			var eventServiceFake = instanceFaker.GetMock<IEventService>()!;

			_ = eventServiceFake.Setup(e => e.Enqueue(It.IsAny<ITaxRateService?>(), ServiceEvents.Inserted, It.IsAny<int>()));
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			await ((ITransactionClient)serviceFunction).Commit();
			/* 
			 * Assert
			 */
			eventServiceFake.Verify(e => e.Enqueue(It.IsAny<ITaxRateService?>(), ServiceEvents.Inserted, It.IsAny<int>()), Times.Exactly(1));
		}

		[TestMethod("Insert commit refreshes cache exactly once")]
		public async Task Insert_Commit_RefreshesCache_ExactlyOnce()
		{
			/* 
			 * Setup
			 */
			var instanceFaker = new InstanceFaker<Insert>();

			var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

			_ = taxRateCacheFake.Setup(e => e.Refresh(It.IsAny<int>()));

			var eventServiceFake = instanceFaker.GetMock<IEventService>()!;

			_ = eventServiceFake.Setup(e => e.Enqueue(It.IsAny<ITaxRateService?>(), ServiceEvents.Inserted, It.IsAny<int>()));
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			await ((ITransactionClient)serviceFunction).Commit();
			/* 
			 * Assert
			 */
			taxRateCacheFake.Verify(e => e.Refresh(It.IsAny<int>()), Times.Exactly(1));
		}

		[TestMethod("Insert commit fires inserted event after notifying cache")]
		public async Task Insert_Commit_InvokedInsertedEvent_AfterNotifyingCache()
		{
			/* 
			 * Setup			
			 * Mockbehavior string ensures no unplanned functions are called on the services.
			 * If you receive an error about function calls not properly set up, and function
			 * calls look ok, it's probably the sequence throwing the exception. This is a
			 * failure in user code, not the test.
			*/
			var sequence = new MockSequence();

			var instanceFaker = new InstanceFaker<Insert>();

			var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

			_ = taxRateCacheFake
				.InSequence(sequence)
				.Setup(e => e.Refresh(It.IsAny<int>()));

			var eventServiceFake = instanceFaker.GetMock<IEventService>()!;

			_ = eventServiceFake
				.InSequence(sequence)
				.Setup(e => e.Enqueue(It.IsAny<ITaxRateService?>(), ServiceEvents.Inserted, It.IsAny<int>()));
			/* 
			 * Test
			 */
			var serviceFunction = instanceFaker.Instance;
			await ((ITransactionClient)serviceFunction).Commit();
			/* 
			 * Assert
			 */
			taxRateCacheFake.Verify(e => e.Refresh(It.IsAny<int>()), Times.Exactly(1));
			eventServiceFake.Verify(e => e.Enqueue(It.IsAny<ITaxRateService?>(), ServiceEvents.Inserted, It.IsAny<int>()));
		}
	}
}