﻿using System.Collections.Immutable;
using System.Data;
using Connected.Common.Types.TaxRates;
using Connected.Entities;
using Connected.ServiceModel;
using Lookup = Connected.Common.Types.TaxRates.TaxRateOps.Lookup;

namespace Common.Types.TaxRates.Ops;


[TestClass]
public class LookupTests
{
	[TestMethod("Lookup invoke returns ImmutableList when set is empty")]
	public async Task Lookup_Invoke_ReturnsImmutableList_WhenEmpty()
	{
		/* 
		 * Setup
		*/
		var dataSet = TaxRateSamples.GetEmptyTaxRateSet();

		var instanceFaker = new InstanceFaker<Lookup>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet!);

		var args = new PrimaryKeyListArgs<int>();
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNotNull(result);
		Assert.IsTrue(result.GetType().IsAssignableToGenericType(typeof(ImmutableList<>)));
	}

	[TestMethod("Lookup invoke returns empty set when argument list is empty")]
	public async Task Lookup_Invoke_ReturnsEmptyImmutableList_WhenArgumentListEmpty()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.CombinedTaxRateSet;

		var instanceFaker = new InstanceFaker<Lookup>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet!);

		var args = new PrimaryKeyListArgs<int>
		{
			IdList = new List<int>()
		};
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNotNull(result);
		Assert.IsTrue(result.Count == 0);
	}

	[TestMethod("Lookup invoke returns empty set when argument list is invalid")]
	public async Task Lookup_Invoke_ReturnsEmptyImmutableList_WhenArgumentListInvalid()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.CombinedTaxRateSet;

		var instanceFaker = new InstanceFaker<Lookup>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet!);

		var args = new PrimaryKeyListArgs<int>();
		/* 
		 * Test
		 */
		var serviceFunction = new Lookup(taxRateCacheFake.Object);
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNotNull(result);
		Assert.IsTrue(result.Count == 0);
	}

	[TestMethod("Lookup invoke returns correct set when argument list is valid")]
	public async Task Lookup_Invoke_ReturnsCorrectSet_WhenArgumentListValid()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.CombinedTaxRateSet;

		var instanceFaker = new InstanceFaker<Lookup>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet!);

		var args = new PrimaryKeyListArgs<int>
		{
			IdList = new[]
			{
				TaxRateSamples.NotModifiedEnabled.Id,
				TaxRateSamples.NewEnabled.Id,
				TaxRateSamples.NotModifiedDisabled.Id
			}
		};

		var expected = new[]
		{
			TaxRateSamples.NotModifiedEnabled,
			TaxRateSamples.NewEnabled,
			TaxRateSamples.NotModifiedDisabled
		};
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNotNull(result);
		Assert.IsTrue(result.SequenceEquivalent(expected));
	}

	[TestMethod("Lookup invoke returns correct set when argument list contains deleted values")]
	public async Task Lookup_Invoke_ReturnsCorrectSet_WhenArgumentListContainsDeletedValues()
	{
		/* 
		 * Setup
		*/
		var dataSet = TaxRateSamples.CombinedTaxRateSet;

		var instanceFaker = new InstanceFaker<Lookup>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet!);

		var args = new PrimaryKeyListArgs<int>
		{
			IdList = new[]
			{
				TaxRateSamples.NotModifiedEnabled.Id,
				TaxRateSamples.DeletedEnabled.Id
			}
		};

		var expected = new[]
		{
			TaxRateSamples.NotModifiedEnabled
		};
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNotNull(result);
		Assert.IsTrue(result.SequenceEquivalent(expected));
	}

	[TestMethod("Lookup invoke returns all entries when no entries are deleted")]
	public async Task Lookup_Invoke_ReturnsAllEntries_WhenNoEntriesAreDeleted()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.GetTaxRates(State.Default)
			.Union(TaxRateSamples.GetTaxRates(State.New));

		var instanceFaker = new InstanceFaker<Lookup>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet!);

		var args = new PrimaryKeyListArgs<int>
		{
			IdList = dataSet.Select(e => e.Id)
		};
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNotNull(result);
		Assert.IsTrue(result.SequenceEqual(dataSet));
	}

	[TestMethod("Lookup invoke returns only non deleted entries when entries are deleted")]
	public async Task Lookup_Invoke_ReturnsNonDeletedOnly_WhenEntriesAreDeleted()
	{
		/* 
		 * Setup
		 */
		var expectedDataSet = TaxRateSamples.GetTaxRates(State.Default)
			.Union(TaxRateSamples.GetTaxRates(State.New));

		var completeDataSet = expectedDataSet.Union(TaxRateSamples.GetTaxRates(State.Deleted));

		var instanceFaker = new InstanceFaker<Lookup>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(completeDataSet);

		var args = new PrimaryKeyListArgs<int>
		{
			IdList = completeDataSet.Select(e => e.Id)
		};
		/* 
		 * Test
		 */
		var serviceFunction = new Lookup(taxRateCacheFake.Object);
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNotNull(result);
		Assert.IsTrue(result.SequenceEqual(expectedDataSet));
	}
}