﻿using Connected.Common.Types.TaxRates;
using Connected.Entities;
using static Connected.Common.Types.TaxRates.TaxRateOps;

namespace Common.Types.TaxRates.Ops;

[TestClass]
public class SelectByRateTests
{
	private static IEnumerable<object[]> TaxRates => TaxRateSamples.CombinedTaxRateSet
		.Select(e => new object[] { e.Rate, e })
		.ToList();

	private static IEnumerable<object[]> DeletedTaxRates => TaxRateSamples.CombinedTaxRateSet
		.Where(e => e.State == State.Deleted)
		.Select(e => new object[] { e.Rate, e })
		.ToList();

	[TestMethod("SelectByRate invoke returns null when rate does not exist")]
	public async Task Select_Invoke_ReturnsNull_WhenRateInvalid()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.CombinedTaxRateSet;

		var instanceFaker = new InstanceFaker<SelectByRate>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet);
		/* 
		 * Ensure id is always out of range 
		 */
		var outOfRangeRate = dataSet.Select(e => e.Rate).DefaultIfEmpty(0).Max() + 1;

		var args = new TaxRateArgs
		{
			Rate = outOfRangeRate
		};
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNull(result);
	}

	[TestMethod("SelectByRate invoke returns null when rate not specified")]
	public async Task Select_Invoke_ReturnsNull_WhenRateNotSpecified()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.CombinedTaxRateSet;

		var instanceFaker = new InstanceFaker<SelectByRate>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet);
		/* 
		 * Ensure id is always out of range 
		 */
		var args = new TaxRateArgs();
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNull(result);
	}

	[TestMethod("SelectByRate invoke returns null when set is empty")]
	public async Task Select_Invoke_ReturnsNull_WhenSetEmpty()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.GetEmptyTaxRateSet();

		var instanceFaker = new InstanceFaker<SelectByRate>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet);

		var args = new TaxRateArgs
		{
			Rate = 1
		};
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNull(result);
	}

	[TestMethod("SelectByRate invoke returns value when rate exists")]
	[DynamicData(nameof(TaxRates))]
	public async Task Select_Invoke_ReturnsValue_WhenRateValid(float rate, ITaxRate expected)
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRates
			.Select(e => e[1] as TaxRate)
			.ToList();

		var instanceFaker = new InstanceFaker<SelectByRate>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet);
		/* 
		 * Ensure id is always out of range 
		 */
		var args = new TaxRateArgs
		{
			Rate = rate
		};
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.AreSame(result, expected);
	}

	[TestMethod("Select invoke returns null value when rate exists but is deleted")]
	[DynamicData(nameof(DeletedTaxRates))]
	public async Task Select_Invoke_ReturnsNull_WhenRateValidAndDeleted(float rate)
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRates
			.Select(e => e[1] as TaxRate)
			.ToList();

		var instanceFaker = new InstanceFaker<SelectByRate>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet);
		/*
		 * Ensure id is always out of range 
		 */
		var args = new TaxRateArgs
		{
			Rate = rate
		};
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNull(result);
	}
}

