﻿using Connected.Common.Types.TaxRates;
using Connected.Entities;
using Connected.ServiceModel;
using static Connected.Common.Types.TaxRates.TaxRateOps;

namespace Common.Types.TaxRates.Ops;

[TestClass]
public class SelectTests
{
	private static IEnumerable<object[]> TaxRates => TaxRateSamples.CombinedTaxRateSet
		.Select(e => new object[] { e.Id, e })
		.ToList();
	private static IEnumerable<object[]> DeletedTaxRates => TaxRateSamples.CombinedTaxRateSet
		.Where(e => e.State == State.Deleted)
		.Select(e => new object[] { e.Id, e })
		.ToList();

	[TestMethod("Select invoke returns null when id does not exist")]
	public async Task Select_Invoke_ReturnsNull_WhenIdInvalid()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.CombinedTaxRateSet;

		var instanceFaker = new InstanceFaker<Select>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		_ = taxRateCacheFake
			.As<IEnumerable<TaxRate>>()
			.Setup(e => e.GetEnumerator())
			.Returns(dataSet.GetEnumerator());
		/* 
		 * Ensure id is always out of range 
		 */
		var outOfRangeId = dataSet.Select(e => e.Id).DefaultIfEmpty(0).Max() + 1;
		var args = new PrimaryKeyArgs<int>(outOfRangeId);
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		*/
		Assert.IsNull(result);
	}

	[TestMethod("Select invoke returns null when set is empty")]
	public async Task Select_Invoke_ReturnsNull_WhenSetEmpty()
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRateSamples.GetEmptyTaxRateSet();

		var instanceFaker = new InstanceFaker<Select>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		_ = taxRateCacheFake
			.As<IEnumerable<TaxRate>>()
			.Setup(e => e.GetEnumerator())
			.Returns(dataSet.GetEnumerator());
		/* 
		 * Ensure id is always out of range 
		 */
		var args = new PrimaryKeyArgs<int>(1);
		/* 
		 * Test
		*/
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.IsNull(result);
	}

	[TestMethod("Select invoke returns value when id exists")]
	[DynamicData(nameof(TaxRates))]
	public async Task Select_Invoke_ReturnsValue_WhenIdValid(int id, ITaxRate expected)
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRates.Select(e => e[1] as TaxRate).ToList();

		var instanceFaker = new InstanceFaker<Select>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable(dataSet!);
		/* 
		 * Ensure id is always out of range 
		 */
		var args = new PrimaryKeyArgs<int>(id);
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.AreSame(result, expected);
	}

	[TestMethod("Select invoke returns deleted value when id exists")]
	[DynamicData(nameof(DeletedTaxRates))]
	public async Task Select_Invoke_ReturnsDeletedValue_WhenIdValid(int id, ITaxRate expected)
	{
		/* 
		 * Setup
		 */
		var dataSet = TaxRates.Select(e => e[1] as TaxRate).ToList();

		var instanceFaker = new InstanceFaker<Select>();

		var taxRateCacheFake = instanceFaker.GetMock<ITaxRateCache>()!;

		taxRateCacheFake.SetupIEnumerable<TaxRate>(dataSet!);
		/* 
		 * Ensure id is always out of range 
		 */
		var args = new PrimaryKeyArgs<int>(id);
		/* 
		 * Test
		 */
		var serviceFunction = instanceFaker.Instance;
		var result = await serviceFunction.Invoke(args);
		/* 
		 * Assert
		 */
		Assert.AreSame(result, expected);
	}
}

